#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QFileDialog>
#include <QDesktopServices>
#include <QMessageBox>
#include <QThread>
#include <QTimer>
#include <QRegExp>
#include <QProcessEnvironment>
#include <QPixmap>
#include <QProcess>
#include <QDebug>

#include "trackerworker.h"
#include "controlworker.h"
#include "crazyflieworker.h"
#include "quaditem.h"
#include "slotprocess.h"
#include "qFlightInstruments.h"
#include "graph_blocks.h"

#include <iostream>
MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    setpointList(new QStandardItemModel(this)),
    nextSpTimer(new QTimer(this)),
    sp_roll(0.0f),
    sp_pitch(0.0f),
    sp_yaw(0.0f),
    sp_thrust(0.0f),
    trackerTimer(new QTimer(this)),
    workerStartTimer(new QTimer(this)),
    crazyflieTimer(new QTimer(this)),
    backendProcess(new QProcess(this)),
    matlabProcess(new QProcess(this)),
    pollGamepadTimer(new QTimer(this)),
    connectedWorkers(0)
{

    ui->setupUi(this);

    QGraphicsScene *posScene = new QGraphicsScene(this);
    // ui->posView->setScene(posScene);
    QuadItem * posIndicator = new QuadItem();
    posScene->addItem(posIndicator);

    QGraphicsScene *attScene = new QGraphicsScene(this);
    //ui->attView->setScene(attScene);
    QADI * attIndicator = new QADI();
    attScene->addWidget(attIndicator);


    workerStartTimer->setSingleShot(true);

    /* Create a thread for workers */
    //QThread* workerThread = new QThread(this);

    /* Create a worker to update the tracker */
   // TrackerWorker * trackerWorker = new TrackerWorker();

    /* Move it to the worker thread. This means that slots of this worker
     * will run in the worker thread, and not block the UI */
    //trackerWorker->moveToThread(workerThread);

    /* Connect tracker worker */
    //connect(trackerWorker, SIGNAL (finished(float, float, float, float, float, float)),
    //        this, SLOT (updateTracker(float, float, float, float, float, float)));

    //connect(trackerWorker, SIGNAL (finished(float, float, float, float, float, float)),
    //        posIndicator, SLOT(updatePos(float,float,float,float,float,float)));

    //connect(trackerWorker, SIGNAL (finished(float, float, float, float, float, float)),
    //        attIndicator, SLOT(updateAttitude(float, float, float, float, float, float)));

    backendProcess->setProcessChannelMode(QProcess::MergedChannels);

    /* Create another worker for the control graph */
    QThread * cwThread = new QThread(this);
    ControlWorker * controlWorker = new ControlWorker();
    controlWorker->moveToThread(cwThread);

    /* Connect signals from control worker */
    connect(controlWorker, SIGNAL (gotNodes(QStringList)), this, SLOT (newNodes(QStringList)));
    connect(controlWorker, SIGNAL (gotParams(QStringList)), this, SLOT (newParams(QStringList)));
    connect(controlWorker, SIGNAL (gotParamValue(QString, QString, float)), this, SLOT (newParamValue(QString, QString, float)));
    connect(controlWorker, SIGNAL (gotNodeOutput(QString, float)), this, SLOT (newNodeOutput(QString, float)));
    connect(controlWorker, SIGNAL (gotConstantBlocks(QStringList)), this, SLOT (newConstantBlocks(QStringList)));
    connect(controlWorker, SIGNAL (graphRendered(QString)), this, SLOT (newControlGraph(QString)));
    connect(controlWorker, SIGNAL (paramSet(QString, QString)), controlWorker, SLOT (getParamValue(QString, QString)));

    /* Signals to control worker */
    // connect(ui->pbControlRefresh, SIGNAL (clicked()), controlWorker, SLOT (getNodes()));
    // connect(ui->nodeSelect, SIGNAL (currentIndexChanged(QString)), controlWorker, SLOT (getParams(QString)));
    connect(this, SIGNAL (getParamValue(QString, QString)), controlWorker, SLOT (getParamValue(QString, QString)));
    connect(this, SIGNAL (setParamValue(QString, QString, float)), controlWorker, SLOT (setParamValue(QString, QString, float)));
    connect(this, SIGNAL (getNodeOutput(QString)), controlWorker, SLOT (getNodeOutput(QString)));

    /* Create another work for the crazyflie communication */
    QThread * cfThread = new QThread(this);
    crazyflieWorker = new CrazyflieWorker();
    crazyflieWorker->moveToThread(cfThread);

    QThread * logThread = new QThread(this);
    logWorker = new LogWorker();
    logWorker->moveToThread(logThread);

    /* connect signals for crazyflie worker */
    connect(ui->logEntriesComboBox1, SIGNAL (currentIndexChanged(int)), this, SLOT (getLogEntryBoxChange1(int)));
    connect(ui->logEntriesComboBox2, SIGNAL (currentIndexChanged(int)), this, SLOT (getLogEntryBoxChange2(int)));
    connect(ui->logEntriesComboBox3, SIGNAL (currentIndexChanged(int)), this, SLOT (getLogEntryBoxChange3(int)));
    connect(ui->logEntriesComboBox4, SIGNAL (currentIndexChanged(int)), this, SLOT (getLogEntryBoxChange4(int)));
    connect(ui->logEntriesComboBox5, SIGNAL (currentIndexChanged(int)), this, SLOT (getLogEntryBoxChange5(int)));
    connect(ui->pb_startLog, SIGNAL (clicked()), logWorker, SLOT (getLogValues(QString, QString, QString, QString, QString)));
    connect(logWorker, SIGNAL (gotLogValues(QStringList)), this, SLOT (graphLogs(QStringList)));

    connect(this, SIGNAL (rateSetpointSignal(float, float, float, float)), crazyflieWorker, SLOT (setCurrAttRateSetpoint(float, float, float, float)));
    connect(this, SIGNAL (angleSetpointSignal(float, float, float, float)), crazyflieWorker, SLOT (setCurrAttSetpoint(float, float, float, float)));
    connect(this, SIGNAL (mixedSetpointSignal(float, float, float, float)), crazyflieWorker, SLOT (setCurrMixedSetpoint(float, float, float, float)));
    connect(crazyflieTimer, SIGNAL(timeout()), this, SLOT(trigger_send_setpoint()));
    connect(this, SIGNAL(triggerAttSetpointSend()), crazyflieWorker, SLOT(sendAttSetpoint()));
    connect(this, SIGNAL(triggerAttRateSetpointSend()), crazyflieWorker, SLOT(sendAttRateSetpoint()));
    connect(this, SIGNAL(triggerMixedAttSetpointSend()), crazyflieWorker, SLOT(sendMixedAttSetpoint()));
    connect(crazyflieWorker, SIGNAL (gotParamGroups(QStringList)), this, SLOT (setParamGroups(QStringList)));
    connect(crazyflieWorker, SIGNAL (gotLogVariables(QStringList)), this, SLOT (setLoggingVariableBox(QStringList)));

    connect(ui->paramGroupComboBox, SIGNAL (currentIndexChanged(int)), this, SLOT (getGroupBoxChange(int)));
    connect(ui->paramGroupComboBox_2, SIGNAL (currentIndexChanged(int)), this, SLOT (getGroupBox2Change(int)));
    connect(ui->paramEntriesComboBox, SIGNAL (currentIndexChanged(int)), this, SLOT (getEntryBoxChange(int)));
    connect(ui->paramEntriesComboBox_2, SIGNAL (currentIndexChanged(int)), this, SLOT (getEntryBoxChange2(int)));
    connect(ui->pb_getParam, SIGNAL (clicked()), this, SLOT (on_pb_getParam_click()));
    connect(ui->pb_setParam, SIGNAL (clicked()), this, SLOT (on_pb_setParam_click()));

    connect(crazyflieWorker, SIGNAL (gotGroupEntries(int, QStringList)), this, SLOT (newGroupEntries(int, QStringList)));
    connect(this, SIGNAL (getGroupEntries(int, QString)), crazyflieWorker, SLOT (getGroupEntries(int, QString)));
    connect(this, SIGNAL (getParamValue(QString)), crazyflieWorker, SLOT (getParamValue(QString)));
    connect(this, SIGNAL (setParamValue(QString, double)), crazyflieWorker, SLOT (setParamValue(QString, double)));
    connect(crazyflieWorker, SIGNAL (gotParamValue(QString, float)), this, SLOT (displayParamValue(QString, float)));

    /* Connect and disconnect from backend when signals emitted */
    //connect(workerStartTimer, SIGNAL (timeout()), trackerWorker, SLOT (connectBackend()));
    connect(workerStartTimer, SIGNAL (timeout()), logWorker, SLOT (connectBackend()));
    connect(workerStartTimer, SIGNAL (timeout()), controlWorker, SLOT (connectBackend()));
    connect(workerStartTimer, SIGNAL (timeout()), crazyflieWorker, SLOT (connectBackend()));

    //connect(this, SIGNAL (connectWorkers()), trackerWorker, SLOT (connectBackend()));
    connect(this, SIGNAL (connectWorkers()), logWorker, SLOT (connectBackend()));
    connect(this, SIGNAL (connectWorkers()), controlWorker, SLOT (connectBackend()));
    connect(this, SIGNAL (connectWorkers()), crazyflieWorker, SLOT (connectBackend()));

    //connect(this, SIGNAL (disconnectWorkers()), trackerWorker, SLOT (disconnectBackend()));
    connect(this, SIGNAL (disconnectWorkers()), logWorker, SLOT (disconnectBackend()));
    connect(this, SIGNAL (disconnectWorkers()), controlWorker, SLOT (disconnectBackend()));
    connect(this, SIGNAL (disconnectWorkers()), crazyflieWorker, SLOT (disconnectBackend()));

    connect(backendProcess, SIGNAL (started()), this, SLOT (backendStarted()));
    connect(backendProcess, SIGNAL (errorOccurred(QProcess::ProcessError)), this, SLOT (backendError(QProcess::ProcessError)));
    connect(backendProcess, SIGNAL (finished(int, QProcess::ExitStatus)), this, SLOT (backendFinished(int, QProcess::ExitStatus)));
    connect(backendProcess, SIGNAL (readyReadStandardOutput()), this, SLOT (backendRead()));
    connect(backendProcess, SIGNAL (readyReadStandardError()), this, SLOT (backendRead()));

    connect(controlWorker, SIGNAL (connected()), this, SLOT (workerConnected()));
    connect(controlWorker, SIGNAL (disconnected()), this, SLOT (workerDisconnected()));

    //connect(trackerWorker, SIGNAL (connected()), this, SLOT (workerConnected()));
    //connect(trackerWorker, SIGNAL (disconnected()), this, SLOT (workerDisconnected()));

    connect(logWorker, SIGNAL (connected()), this, SLOT (workerConnected()));
    connect(logWorker, SIGNAL (disconnected()), this, SLOT (workerDisconnected()));

    connect(crazyflieWorker, SIGNAL (connected()), this, SLOT (workerConnected()));
    connect(crazyflieWorker, SIGNAL (disconnected()), this, SLOT (workerDisconnected()));

    /* Connect refresh button and refresh timer to tracker worker */
    connect(trackerTimer, SIGNAL(timeout()), this, SLOT(updatePosAtt()));
    // connect(ui->pbRefresh, SIGNAL (clicked()), this, SLOT (updatePosAtt()));

    //connect(this, SIGNAL(getPosAttFromBackend()), trackerWorker, SLOT(process()));

    /* Timer used for next setpoint */
    nextSpTimer->setSingleShot(true);
    connect(nextSpTimer, SIGNAL (timeout()), this, SLOT (on_pbNextSetpoint_clicked()));

    /* Start the things */
    trackerTimer->start(100);
    crazyflieTimer->start(100);
    //workerThread->start();
    cwThread->start();
    cfThread->start();
    logThread->start();
    matlabProcess->start();

    /* Connect the setpointlist to the model */
    // ui->setpointList->setModel(setpointList);

    ui->pb_getParam->setEnabled(false);
    ui->pb_setParam->setEnabled(false);
    ui->setValueSpin->setMaximum(10000.00);
    ui->setValueSpin->setMinimum(-10000.00);
    ui->setValueSpin->setSingleStep(0.01);
    ui->setValueSpin->setValue(0);

    /* Connect various things that can result in sending setpoints */
    // connect(ui->pbSendSetpoint, SIGNAL (clicked()), this, SLOT (sendSetpoints()));
    // connect(ui->setpointList, SIGNAL (doubleClicked(QModelIndex)), this, SLOT (sendSelectedSetpoint()));

    /* start gamepad monitor */
    gamepadMonitor = new GamepadMonitor();
    if(gamepadMonitor->isGamepadConnected()){
        onGamepadConnect();
    }else{
        onGamepadDisconnect();
    }

    connect(gamepadMonitor, SIGNAL(gamepadDisconnected()), this, SLOT(onGamepadDisconnect()));
    connect(gamepadMonitor, SIGNAL(gamepadConnected()), this, SLOT(onGamepadConnect()));
    connect(ui->pb_configRoll, SIGNAL(clicked()), this, SLOT(on_pb_configRoll_clicked()));
    connect(pollGamepadTimer, SIGNAL(timeout()), this, SLOT(updateGamepad()));
    pollGamepadTimer->start(50);

    //gamepad configure tab setup
    ui->gamepadRollScale->setValue((int)gamepadMonitor->getRollScale());
    ui->gamepadPitchScale->setValue((int)gamepadMonitor->getPitchScale());
    ui->gamepadYawScale->setValue((int)gamepadMonitor->getYawScale());
    ui->gamepadThrustScale->setValue((int)gamepadMonitor->getThrustScale());



    /* Populate scripts list */
    QDir scriptsDir("scripts/");
    QStringList scripts = scriptsDir.entryList();
    for (int i = 0; i < scripts.size(); i++) {
        QAction * action = ui->menuScripts->addAction(scripts[i]);
        SlotProcess * scriptProcess = new SlotProcess(action);
        scriptProcess->setProgram(scriptsDir.filePath(scripts[i]));
        connect(action, SIGNAL (triggered()), scriptProcess, SLOT (startProcess()));
    }
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::connectMatlab()
{

}

void MainWindow::on_run_data_analysis_button_clicked()
{
    QProcess * matProcess = new QProcess(this);
    QString matlab = "/usr/local/bin/matlab";
    matProcess->start(matlab, QStringList() << QString("-nodesktop")<<  QString("-nosplash")<<QString("-r")<<QString("run('../../../controls/DataAnalysisTool/Tool/GUI.m');"));
}

void MainWindow::updatePosAtt()
{
    // emit(getNodeOutput(ui->xPositionSelect->currentText()));
    // emit(getNodeOutput(ui->yPositionSelect->currentText()));
    // emit(getNodeOutput(ui->zPositionSelect->currentText()));
    // emit(getNodeOutput(ui->pAttitudeSelect->currentText()));
    // emit(getNodeOutput(ui->rAttitudeSelect->currentText()));
    // emit(getNodeOutput(ui->yAttitudeSelect->currentText()));
}

void MainWindow::updateTracker(float x, float y, float z, float p, float r, float yaw)
{
    // ui->xActual->setText(QString::number(x));
    // ui->yActual->setText(QString::number(y));
    // ui->zActual->setText(QString::number(z));
    ui->currPitchActual->setText(QString::number(p));
    ui->currRollActual->setText(QString::number(r));
    ui->currYawActual->setText(QString::number(yaw));

    //float dist = sqrt(pow(x - sp_x, 2.0) + pow(y - sp_y, 2.0) + pow(z - sp_z, 2.0));

    // ui->dist->setText(QString::number(dist));

    // if (!nextSpTimer->isActive() && ui->autonavEnabled->isChecked() &&
    //         (dist < ui->autonavThreshold->text().toFloat())) {
    //     nextSpTimer->start(ui->autonavDelay->text().toInt());
    // }
}

void MainWindow::on_pbStart_clicked()
{
    backendProcess->setProgram(ui->backendPath->text());

    backendProcess->start();
}

void MainWindow::backendStarted()
{
    ui->pbStart->setEnabled(false);
    ui->pbStop->setEnabled(true);
    workerStartTimer->start(750);
}

void MainWindow::backendError(QProcess::ProcessError error)
{
    ui->vConsole->append(backendProcess->errorString());
}

void MainWindow::backendRead()
{
    ui->vConsole->append(backendProcess->readAll());
}

void MainWindow::workerConnected()
{
    connectedWorkers++;
     if (connectedWorkers == 2) {
        ui->pbStart->setEnabled(false);
        ui->pbConnect->setEnabled(false);
        ui->pbStop->setEnabled(true);
    }
}

void MainWindow::workerDisconnected()
{
    connectedWorkers--;
    if (( connectedWorkers == 0) && (backendProcess->state() == QProcess::Running)) {
        backendProcess->terminate();
    } else if (connectedWorkers == 0) {
        ui->pbStart->setEnabled(true);
        ui->pbConnect->setEnabled(true);
        ui->pbStop->setEnabled(false);
    }
}

void MainWindow::backendFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    ui->pbStart->setEnabled(true);
    ui->pbConnect->setEnabled(true);
    ui->pbStop->setEnabled(false);
}

void MainWindow::on_pbConnect_clicked()
{
    emit(connectWorkers());
}

void MainWindow::on_pbStop_clicked()
{
    emit(disconnectWorkers());
}

void MainWindow::on_chooseBackend_clicked()
{
    QString backendPath = QFileDialog::getOpenFileName(this,
         tr("Path to Backend Executable"));
    ui->backendPath->setText(backendPath);
}


void MainWindow::newNodes(QStringList blocks)
{
    // QComboBox * select = ui->nodeSelect;
    // select->clear();
    // select->addItems(blocks);

    // this->ui->noGraphWarning1->setVisible(false);
    // this->ui->noGraphWarning2->setVisible(false);
    // this->ui->noGraphWarningLine->setVisible(false);
}

void MainWindow::setParamGroups(QStringList groupNames) {
    ui->paramGroupComboBox->addItem("");
    ui->paramGroupComboBox->addItems(groupNames);
    ui->paramGroupComboBox_2->addItem("");
    ui->paramGroupComboBox_2->addItems(groupNames);
}

void MainWindow::setLoggingVariableBox(QStringList variableNames) {
    ui->logVariableList->addItems(variableNames);
}

void MainWindow::getGroupBoxChange(int index) {
    // for(int i = 0; i < ui->paramGroupComboBox->count(); i++) {
    //     ui->paramEntriesComboBox->removeItem(i);
    // }
    ui->paramEntriesComboBox->clear();
    QString qStrGroup = ui->paramGroupComboBox->itemText(index);
    std::string check = qStrGroup.toStdString();
    ui->valueVallabel->clear();
    if(!check.compare("")) {
        ui->pb_getParam->setEnabled(false);
    }
    else {
        emit(getGroupEntries(1, qStrGroup));
    }
}

void MainWindow::getGroupBox2Change(int index) {
    // for(int i = 0; i < ui->paramGroupComboBox->count(); i++) {
    //     ui->paramEntriesComboBox->removeItem(i);
    // }
    ui->paramEntriesComboBox_2->clear();
    QString qStrGroup = ui->paramGroupComboBox_2->itemText(index);
    std::string check = qStrGroup.toStdString();
    if(!check.compare("")) {
        ui->pb_setParam->setEnabled(false);
    }
    else {
        emit(getGroupEntries(2, qStrGroup));
    }
}

void MainWindow::getEntryBoxChange(int index) {
    QString qStrGroup = ui->paramEntriesComboBox->itemText(index);
    std::string check = qStrGroup.toStdString();
    ui->valueVallabel->clear();
    if(!check.compare("")) {
        ui->pb_getParam->setEnabled(false);
    }
    else {
        ui->pb_getParam->setEnabled(true);
    }
}

void MainWindow::getEntryBoxChange2(int index) {
    QString qStrGroup = ui->paramEntriesComboBox_2->itemText(index);
    std::string check = qStrGroup.toStdString();
    if(!check.compare("")) {
        ui->pb_setParam->setEnabled(false);
    }
    else {
        ui->pb_setParam->setEnabled(true);
    }
}

void MainWindow::newGroupEntries(int box, QStringList entries) {
    if(box == 1) {
        ui->paramEntriesComboBox->addItem("");
        ui->paramEntriesComboBox->addItems(entries);
    }
    else {
        ui->paramEntriesComboBox_2->addItem("");
        ui->paramEntriesComboBox_2->addItems(entries);
    }    
}

void MainWindow::on_pb_getParam_click() {
    ui->valueVallabel->clear();
    QString group = ui->paramGroupComboBox->currentText();
    QString entry = ui->paramEntriesComboBox->currentText();
    QString key = group + "." + entry;
    emit(getParamValue(key));
}

void MainWindow::on_pb_setParam_click() {
    QString group = ui->paramGroupComboBox_2->currentText();
    QString entry = ui->paramEntriesComboBox_2->currentText();
    QString key = group + "." + entry;
    double value = ui->setValueSpin->value();
    emit(setParamValue(key, value));
}

void MainWindow::displayParamValue(QString key, float value) {
    ui->valueVallabel->setNum(value);
}

void MainWindow::newConstantBlocks(QStringList blocks)
{
    // ui->xSetpointSelect->clear();
    // ui->xSetpointSelect->addItems(blocks);

    // ui->ySetpointSelect->clear();
    // ui->ySetpointSelect->addItems(blocks);

    // ui->zSetpointSelect->clear();
    // ui->zSetpointSelect->addItems(blocks);

    // ui->yawSetpointSelect->clear();
    // ui->yawSetpointSelect->addItems(blocks);

    // ui->xPositionSelect->clear();
    // ui->xPositionSelect->addItems(blocks);

    // ui->yPositionSelect->clear();
    // ui->yPositionSelect->addItems(blocks);

    // ui->zPositionSelect->clear();
    // ui->zPositionSelect->addItems(blocks);

    // ui->pAttitudeSelect->clear();
    // ui->pAttitudeSelect->addItems(blocks);

    // ui->rAttitudeSelect->clear();
    // ui->rAttitudeSelect->addItems(blocks);

    // ui->yAttitudeSelect->clear();
    // ui->yAttitudeSelect->addItems(blocks);

    // for (ssize_t i = 0; i < blocks.size(); i++) {
    //     if (blocks[i].contains("setpoint", Qt::CaseInsensitive) || blocks[i].contains("sp", Qt::CaseInsensitive)) {
    //         if (blocks[i].contains("x ", Qt::CaseInsensitive)) {
    //             ui->xSetpointSelect->setCurrentIndex(i);
    //         }
    //         if (blocks[i].contains("y ", Qt::CaseInsensitive)) {
    //             ui->ySetpointSelect->setCurrentIndex(i);
    //         }
    //         if (blocks[i].contains("z ", Qt::CaseInsensitive) || blocks[i].contains("alt", Qt::CaseInsensitive)) {
    //             ui->zSetpointSelect->setCurrentIndex(i);
    //         }
    //         if (blocks[i].contains("yaw", Qt::CaseInsensitive) || blocks[i].contains("alt", Qt::CaseInsensitive)) {
    //             ui->yawSetpointSelect->setCurrentIndex(i);
    //         }
    //     }

    //     if (blocks[i] == QString("Pitch")) {
    //         ui->pAttitudeSelect->setCurrentIndex(i);
    //     }

    //     if (blocks[i] == QString("Roll")) {
    //         ui->rAttitudeSelect->setCurrentIndex(i);
    //     }

    //     if (blocks[i] == QString("Yaw")) {
    //         ui->yAttitudeSelect->setCurrentIndex(i);
    //     }

    //     if (blocks[i] == QString("Lidar")) {
    //         ui->zPositionSelect->setCurrentIndex(i);
    //     }

    //     if (blocks[i] == QString("OF X Trim Add")) {
    //         ui->xPositionSelect->setCurrentIndex(i);
    //     }

    //     if (blocks[i] == QString("OF Y Trim Add")) {
    //         ui->yPositionSelect->setCurrentIndex(i);
    //     }
    // }
}

void MainWindow::newNodeOutput(QString node, float val)
{
    // /* Update the nav page if quad is set as the source for pos/att */
    // if (node == ui->xPositionSelect->currentText()) {
    //     ui->xActual->setText(QString::number(val));
    // } else if (node == ui->yPositionSelect->currentText()) {
    //     ui->yActual->setText(QString::number(val));
    // } else if (node == ui->zPositionSelect->currentText()) {
    //     ui->zActual->setText(QString::number(val));
    // } else if (node == ui->pAttitudeSelect->currentText()) {
    //     ui->pitchActual->setText(QString::number(val));
    // } else if (node == ui->rAttitudeSelect->currentText()) {
    //     ui->rollActual->setText(QString::number(val));
    // } else if (node == ui->yAttitudeSelect->currentText()) {
    //     ui->yawActual->setText(QString::number(val));
    // }
}

void MainWindow::newParams(QStringList params)
{
    // QComboBox * select = ui->paramSelect;
    // select->clear();
    // select->addItems(params);
}

void MainWindow::newParamValue(QString node, QString param, float val)
{
    // ui->paramValue->setText(QString::number(val));

    // /* Update the nav page setpoints if it's a setpoint paramvalue */
    // if (node == ui->xSetpointSelect->currentText()) {
    //     ui->xSetpoint->setText(QString::number(val));
    // } else if (node == ui->ySetpointSelect->currentText()) {
    //     ui->ySetpoint->setText(QString::number(val));
    // } else if (node == ui->zSetpointSelect->currentText()) {
    //     ui->zSetpoint->setText(QString::number(val));
    // }
}

void MainWindow::on_paramSelect_currentIndexChanged(const QString &arg1)
{
    // emit(getParamValue(ui->nodeSelect->currentText(), arg1));
}

void MainWindow::on_paramValue_returnPressed()
{
    // emit (setParamValue(ui->nodeSelect->currentText(),
    //                     ui->paramSelect->currentText(),
    //                     ui->paramValue->text().toFloat()));
}


void MainWindow::sendSetpoints()
{

    // if (ui->sendX->isChecked()) {
    //     sp_x = ui->xSetpoint->text().toFloat();
    //     emit (setParamValue(ui->xSetpointSelect->currentText(),
    //                         blockDefs[BLOCK_CONSTANT]->param_names[0], sp_x));
    // }

    // if (ui->sendY->isChecked()) {
    //     sp_y = ui->ySetpoint->text().toFloat();
    //     emit (setParamValue(ui->ySetpointSelect->currentText(),
    //                         blockDefs[BLOCK_CONSTANT]->param_names[0], sp_y));
    // }

    // if (ui->sendZ->isChecked()) {
    //     sp_z = ui->zSetpoint->text().toFloat();
    //     emit (setParamValue(ui->zSetpointSelect->currentText(),
    //                         blockDefs[BLOCK_CONSTANT]->param_names[0], sp_z));
    // }

    // if (ui->sendYaw->isChecked()) {
    //     emit (setParamValue(ui->yawSetpointSelect->currentText(),
    //                         blockDefs[BLOCK_CONSTANT]->param_names[0],
    //                         ui->yawSetpoint->text().toFloat()));
    // }
}

void MainWindow::on_pbAppendSetpoint_clicked()
{
    // QString str("[" + ui->xSetpoint->text() + ", "+
    //             ui->ySetpoint->text() + ", " +
    //             ui->zSetpoint->text() + ", " +
    //         ui->yawSetpoint->text() + "]");

    // setpointList->appendRow(new QStandardItem(str));
}

void MainWindow::on_pbNextSetpoint_clicked()
{
    // QListView * listView = ui->setpointList;
    // if (listView->currentIndex().isValid() && setpointList->index(listView->currentIndex().row() + 1, 0).isValid()) {
    //     listView->setCurrentIndex(setpointList->index(listView->currentIndex().row() + 1, 0));
    // } else {
    //     listView->setCurrentIndex(setpointList->index(0, 0));
    // }
    // sendSelectedSetpoint();
}


void MainWindow::sendSelectedSetpoint()
{
    // if (ui->setpointList->currentIndex().isValid()) {
    //     QRegExp regex("\\[(.*), (.*), (.*), (.*)\\]");
    //     int row = ui->setpointList->currentIndex().row();

    //     regex.indexIn(setpointList->item(row)->text());
    //     ui->xSetpoint->setText(regex.cap(1));
    //     ui->ySetpoint->setText(regex.cap(2));
    //     ui->zSetpoint->setText(regex.cap(3));
    //     ui->yawSetpoint->setText(regex.cap(4));

    //     sendSetpoints();
    // }
}

void MainWindow::on_pbActualToSetpoint_clicked()
{
    // ui->xSetpoint->setText(ui->xActual->text());
    // ui->ySetpoint->setText(ui->yActual->text());
    // ui->zSetpoint->setText(ui->zActual->text());
    // ui->yawSetpoint->setText(ui->yawActual->text());
}

void MainWindow::on_pbDeleteSetpoint_clicked()
{
    // if (ui->setpointList->currentIndex().isValid()) {
    //     setpointList->removeRow(ui->setpointList->currentIndex().row());
    // }
}

// void MainWindow::newControlGraph(QString graph)
// {
//     ui->graphImage->setPixmap(QPixmap(graph));
// }

void MainWindow::on_pbActualToWaypoint_clicked()
{
    // QString str("[" + ui->xActual->text() + ", "+
    //             ui->yActual->text() + ", " +
    //             ui->zActual->text() + ", " +
    //             ui->yawActual->text() + "]");

    // setpointList->appendRow(new QStandardItem(str));
}

void MainWindow::on_pbMoveUp_clicked()
{
    // if (ui->setpointList->currentIndex().isValid()) {
    //     int current = ui->setpointList->currentIndex().row();
    //     if (current > 0) {
    //         setpointList->insertRow(current - 1, setpointList->takeItem(current));
    //         setpointList->removeRow(current + 1);
    //     }
    // }
}

void MainWindow::on_pbMoveDown_clicked()
{
    // if (ui->setpointList->currentIndex().isValid()) {
    //     int current = ui->setpointList->currentIndex().row();
    //     if (current < (setpointList->rowCount() - 1)) {
    //         setpointList->insertRow(current + 2, setpointList->takeItem(current));
    //         setpointList->removeRow(current);
    //     }
    // }
}

void MainWindow::on_pbInsertSetpoint_clicked()
{
    // int current = 0;
    // if (ui->setpointList->currentIndex().isValid()) {
    //     current = ui->setpointList->currentIndex().row();
    // }

    // QString str("[" + ui->xSetpoint->text() + ", "+
    //             ui->ySetpoint->text() + ", " +
    //             ui->zSetpoint->text() + ", " +
    //             ui->yawSetpoint->text() + "]");
    // setpointList->insertRow(current, new QStandardItem(str));
}

void MainWindow::on_pbSaveWaypoints_clicked()
{
    QString filter = "Waypoints (*.wpt)";
    QString savePath = QFileDialog::getSaveFileName(this, tr("Save waypoint"), QString(), filter, &filter, QFileDialog::DontUseNativeDialog);

    if (!savePath.isEmpty()) {
        QFileInfo finfo(savePath);
        if (finfo.suffix().isNull())
            savePath.append(".wpt");
        QFile f(savePath);
        f.open(QIODevice::WriteOnly | QIODevice::Text);
        for (int i= 0; i < setpointList->rowCount(); i++) {
            f.write(setpointList->item(i)->text().toStdString().c_str());
            f.write("\n");
        }
        f.close();
    }
}

void MainWindow::on_pbLoadWaypoints_clicked()
{

    QString filter = "Waypoints (*.wpt)";
    QString openPath = QFileDialog::getOpenFileName(this, tr("Open Waypoint File"), QString(),filter);

    if (!openPath.isEmpty()) {
        QFile f(openPath);
        f.open(QIODevice::ReadOnly | QIODevice::Text);

        setpointList->clear();
        while (!f.atEnd()) {
            QString line = f.readLine();
            line.chop(1);
            setpointList->appendRow(new QStandardItem(line));
        }
        f.close();
    }
}

void MainWindow::on_xSetpoint_returnPressed()
{
    // sp_x = ui->xSetpoint->text().toFloat();
    // emit (setParamValue(ui->xSetpointSelect->currentText(),
    //                     blockDefs[BLOCK_CONSTANT]->param_names[0], sp_x));
}

void MainWindow::on_ySetpoint_returnPressed()
{
    // sp_y = ui->ySetpoint->text().toFloat();
    // emit (setParamValue(ui->ySetpointSelect->currentText(),
    //                     blockDefs[BLOCK_CONSTANT]->param_names[0], sp_y));
}

void MainWindow::on_zSetpoint_returnPressed()
{
    // sp_z = ui->zSetpoint->text().toFloat();
    // emit (setParamValue(ui->zSetpointSelect->currentText(),
    //                     blockDefs[BLOCK_CONSTANT]->param_names[0], sp_z));
}

void MainWindow::on_yawSetpoint_returnPressed()
{
    // emit (setParamValue(ui->yawSetpointSelect->currentText(),
    //                     blockDefs[BLOCK_CONSTANT]->param_names[0],
    //         ui->yawSetpoint->text().toFloat()));
}

void MainWindow::on_applySetpointButton_clicked()
{
    sp_roll   = ui->rollSetpointBox->text().toFloat();
    sp_pitch  = ui->pitchSetpointBox->text().toFloat();
    sp_yaw    = ui->yawSetpointBox->text().toFloat();
    sp_thrust = ui->tActual -> value();

    if(ui->angleRadioButton->isChecked()) {
        //send as angle setpoint
        emit(angleSetpointSignal(sp_roll, sp_pitch, sp_yaw, sp_thrust));
    }
    else if (ui->rateRadioButton->isChecked()) {
        //send as rate setpoint
        emit(rateSetpointSignal(sp_roll, sp_pitch, sp_yaw, sp_thrust));
    }


}

void MainWindow::on_stopSetpointButton_clicked()
{
    sp_roll   = 0.0;
    sp_pitch  = 0.0;
    sp_yaw    = 0.0;
    sp_thrust = 0.0;

    if(ui->angleRadioButton->isChecked()) {
        //send as angle setpoint
        emit(angleSetpointSignal(sp_roll, sp_pitch, sp_yaw, sp_thrust));
    }
    else if (ui->rateRadioButton->isChecked()) {
        //send as rate setpoint
        emit(rateSetpointSignal(sp_roll, sp_pitch, sp_yaw, sp_thrust));
    }
}
void MainWindow::on_tActual_sliderMoved(int position)
{
    QToolTip::showText(QCursor::pos(), QString::number(position), nullptr);

}


/******** Gamepad handlers ********/

void MainWindow::onGamepadDisconnect(){

    emit(mixedSetpointSignal(0, 0, 0, 0));
    emit(rateSetpointSignal(0, 0, 0, 0));
    emit(angleSetpointSignal(0, 0, 0, 0));
   //navigation tab
    ui->pitchSetpointBox->setText("0");
    ui->rollSetpointBox->setText("0");
    ui->yawSetpointBox->setText("0");
    ui->tActual->setValue(0);
    ui->rbManualSetpoint->setChecked(true);

    //gamepad tab
    ui->noGamepadWarning->show();
    ui->rbGamepadControl->setEnabled(false);
    ui->pb_configPitch->setEnabled(false);
    ui->pb_configRoll->setEnabled(false);
    ui->pb_configYaw->setEnabled(false);
    ui->pb_configThrust->setEnabled(false);
    ui->pb_resetConfig->setEnabled(false);
    ui->gamepadRollScale->setEnabled(false);
    ui->rollScaleLabel->setEnabled(false);
    ui->gamepadPitchScale->setEnabled(false);
    ui->pitchScaleLabel->setEnabled(false);
    ui->gamepadYawScale->setEnabled(false);
    ui->yawScaleLabel->setEnabled(false);
    ui->gamepadThrustScale->setEnabled(false);
    ui->thrustScaleLabel->setEnabled(false);
    ui->thrustScaleNote->setEnabled(false);



}

void MainWindow::onGamepadConnect(){

    //navigation tab
    ui->rbGamepadControl->setEnabled(true);

    //gamepad tab
    ui->noGamepadWarning->hide();
    ui->pb_configPitch->setEnabled(true);
    ui->pb_configRoll->setEnabled(true);
    ui->pb_configYaw->setEnabled(true);
    ui->pb_configThrust->setEnabled(true);
    ui->pb_resetConfig->setEnabled(true);

    ui->gamepadRollScale->setEnabled(true);
    ui->rollScaleLabel->setEnabled(true);
    ui->gamepadPitchScale->setEnabled(true);
    ui->pitchScaleLabel->setEnabled(true);
    ui->gamepadYawScale->setEnabled(true);
    ui->yawScaleLabel->setEnabled(true);
    ui->gamepadThrustScale->setEnabled(true);
    ui->thrustScaleLabel->setEnabled(true);
    ui->thrustScaleNote->setEnabled(true);
}

void MainWindow::updateGamepad(){
    if(ui->tabWidget->currentWidget() == ui->Gamepad){
        //on gamepad tab
        ui->rollVizBar->setValue((int) gamepadMonitor->getRoll());
        ui->yawVizBar->setValue((int) gamepadMonitor->getYaw());
        ui->pitchVizBar->setValue((int) gamepadMonitor->getPitch());
        ui->thrustVizBar->setValue((int) gamepadMonitor->getThrust());
    }else if(ui->tabWidget->currentWidget() == ui->navigation){
        if(ui->rbGamepadControl->isChecked()){
            float roll = gamepadMonitor->getRoll();
            float pitch = gamepadMonitor->getPitch();
            float yaw = gamepadMonitor->getYaw();
            float thrust = gamepadMonitor->getThrust();

            //create mixed setpoint
            emit(mixedSetpointSignal(roll, pitch, yaw, thrust));

            ui->rollSetpointBox->setText(QString::number(roll));
            ui->pitchSetpointBox->setText(QString::number(pitch));
            ui->yawSetpointBox->setText(QString::number(yaw));
            ui->tActual->setValue((int) thrust);
        }
    }
}

void MainWindow::on_pb_resetConfig_clicked()
{
    gamepadMonitor->resetConfig();
    ui->gamepadRollScale->setValue((int)gamepadMonitor->getRollScale());
    ui->gamepadPitchScale->setValue((int)gamepadMonitor->getPitchScale());
    ui->gamepadYawScale->setValue((int)gamepadMonitor->getYawScale());
    ui->gamepadThrustScale->setValue((int)gamepadMonitor->getThrustScale());
}

void MainWindow::on_pb_configRoll_clicked()
{
    gamepadMonitor->configureAxis("roll");
}

void MainWindow::on_pb_configYaw_clicked()
{
    gamepadMonitor->configureAxis("yaw");
}

void MainWindow::on_pb_configPitch_clicked()
{
    gamepadMonitor->configureAxis("pitch");
}

void MainWindow::on_pb_configThrust_clicked()
{
    gamepadMonitor->configureAxis("thrust");
}



void MainWindow::on_rbManualSetpoint_toggled(bool checked)
{
    if(checked) {
        emit(mixedSetpointSignal(0, 0, 0, 0));
        emit(rateSetpointSignal(0, 0, 0, 0));
        emit(angleSetpointSignal(0, 0, 0, 0));
        
        ui->pitchSetpointBox->setText("0");
        ui->rollSetpointBox->setText("0");
        ui->yawSetpointBox->setText("0");
        ui->tActual->setValue(0);
        ui->rollSetpointBox-> setEnabled(true);
        ui->pitchSetpointBox-> setEnabled(true);
        ui->yawSetpointBox-> setEnabled(true);
        ui->tActual-> setEnabled(true);
        ui->angleRadioButton-> setEnabled(true);
        ui->rateRadioButton-> setEnabled(true);
        ui->applySetpointButton->setEnabled(true);
        ui->stopSetpointButton->setEnabled(true);
    }
}

void MainWindow::on_rbGamepadControl_toggled(bool checked)
{

    if(checked){
        ui->rollSetpointBox-> setEnabled(false);
        ui->pitchSetpointBox-> setEnabled(false);
        ui->yawSetpointBox-> setEnabled(false);
        ui->tActual-> setEnabled(false);
        ui->angleRadioButton-> setEnabled(false);
        ui->rateRadioButton-> setEnabled(false);
        ui->applySetpointButton->setEnabled(false);
        ui->stopSetpointButton->setEnabled(false);
    }
}

void MainWindow::trigger_send_setpoint(){
    if(ui->rbGamepadControl->isChecked()){
        //using gamepad, send mixed attitude setpoint
        emit(triggerMixedAttSetpointSend());
    }else{
        if(ui->angleRadioButton->isChecked()){
            emit(triggerAttSetpointSend());
        }else{
            emit(triggerAttRateSetpointSend());
        }
    }
}

void MainWindow::on_gamepadYawScale_valueChanged(int arg1)
{
    gamepadMonitor->setYawScale(arg1);
    ui->yawVizBar->setMinimum(std::abs(arg1) * -1);
    ui->yawVizBar->setMaximum(std::abs(arg1));
}

void MainWindow::on_gamepadThrustScale_valueChanged(int arg1)
{
    gamepadMonitor->setThrustScale(arg1);
    //Dont need to change the scale of the thrust viz bar because it is always interpreted as 0 to 60,000
}

void MainWindow::on_gamepadRollScale_valueChanged(int arg1)
{
    gamepadMonitor->setRollScale(arg1);
    ui->rollVizBar->setMinimum(std::abs(arg1) * -1);
    ui->rollVizBar->setMaximum(std::abs(arg1));
}

void MainWindow::on_gamepadPitchScale_valueChanged(int arg1)
{
    gamepadMonitor->setPitchScale(arg1);
    ui->pitchVizBar->setMinimum(std::abs(arg1) * -1);
    ui->pitchVizBar->setMaximum(std::abs(arg1));
}

void MainWindow::graphLogs(QStringList logs)
{   
    ui->dataPlot->graph(0)->setPen(QPen(Qt::blue)); 
    ui->dataPlot->graph(1)->setPen(QPen(Qt::red)); 
    ui->dataPlot->graph(2)->setPen(QPen(Qt::green)); 
    ui->dataPlot->graph(3)->setPen(QPen(Qt::black)); 
    //ui->dataPlot->graph(4)->setPen(QPen(Qt::purple)); 


    //parse data and put into vector
    QVector<float> x0(logs.size()), y0(logs.size());
    QVector<float> x1(logs.size()), y1(logs.size());
    QVector<float> x2(logs.size()), y2(logs.size());
    QVector<float> x3(logs.size()), y3(logs.size());
    QVector<float> x4(logs.size()), y4(logs.size());

    for(int i=0; i< logs.size(); i++) {
        //use tokenizer here
    }

    //start graph and set data
    ui->dataPlot->addGraph();
   // ui->dataPlot->graph(0)->setData(x0, y0);

    //set lables and boundaries
    ui->dataPlot->xAxis->setLabel("time");
    ui->dataPlot->yAxis->setLabel("angle");

    ui->dataPlot->xAxis->setRange(0, 1);
    ui->dataPlot->yAxis->setRange(0, 360);

    ui->dataPlot->replot();
}

